from typing import Optional
from collections import UserDict
from edsl.jobs.buckets.TokenBucket import TokenBucket
from edsl.jobs.buckets.ModelBuckets import ModelBuckets

# from functools import wraps
from threading import RLock

from edsl.jobs.decorators import synchronized_class


@synchronized_class
class BucketCollection(UserDict):
    """A Jobs object will have a whole collection of model buckets, as multiple models could be used.

    The keys here are the models, and the values are the ModelBuckets objects.
    Models themselves are hashable, so this works.
    """

    def __init__(self, infinity_buckets: bool = False):
        """Create a new BucketCollection.
        An infinity bucket is a bucket that never runs out of tokens or requests.
        """
        super().__init__()
        self.infinity_buckets = infinity_buckets
        self.models_to_services = {}
        self.services_to_buckets = {}
        self._lock = RLock()

        from edsl.config import CONFIG
        import os

        url = os.environ.get("EDSL_REMOTE_TOKEN_BUCKET_URL", None)

        if url == "None" or url is None:
            self.remote_url = None
            # print(f"Using remote token bucket URL: {url}")
        else:
            self.remote_url = url

    @classmethod
    def from_models(
        cls, models_list: list, infinity_buckets: bool = False
    ) -> "BucketCollection":
        """Create a BucketCollection from a list of models."""
        bucket_collection = cls(infinity_buckets=infinity_buckets)
        for model in models_list:
            bucket_collection.add_model(model)
        return bucket_collection

    def get_tokens(
        self, model: "LanguageModel", bucket_type: str, num_tokens: int
    ) -> int:
        """Get the number of tokens remaining in the bucket."""
        relevant_bucket = getattr(self[model], bucket_type)
        return relevant_bucket.get_tokens(num_tokens)

    def __repr__(self):
        return f"BucketCollection({self.data})"

    def add_model(self, model: "LanguageModel") -> None:
        """Adds a model to the bucket collection.

        This will create the token and request buckets for the model."""

        # compute the TPS and RPS from the model
        if not self.infinity_buckets:
            TPS = model.tpm / 60.0
            RPS = model.rpm / 60.0
        else:
            TPS = float("inf")
            RPS = float("inf")

        if model.model not in self.models_to_services:
            service = model._inference_service_
            if service not in self.services_to_buckets:
                requests_bucket = TokenBucket(
                    bucket_name=service,
                    bucket_type="requests",
                    capacity=RPS,
                    refill_rate=RPS,
                    remote_url=self.remote_url,
                )
                tokens_bucket = TokenBucket(
                    bucket_name=service,
                    bucket_type="tokens",
                    capacity=TPS,
                    refill_rate=TPS,
                    remote_url=self.remote_url,
                )
                self.services_to_buckets[service] = ModelBuckets(
                    requests_bucket, tokens_bucket
                )
            self.models_to_services[model.model] = service
            self[model] = self.services_to_buckets[service]
        else:
            self[model] = self.services_to_buckets[self.models_to_services[model.model]]

    def visualize(self) -> dict:
        """Visualize the token and request buckets for each model."""
        plots = {}
        for model in self:
            plots[model] = self[model].visualize()
        return plots



"""A list of Agents
"""

from __future__ import annotations
import csv
import sys
from collections import UserList
from collections.abc import Iterable

from typing import Any, List, Optional, Union, TYPE_CHECKING

from simpleeval import EvalWithCompoundTypes, NameNotDefined

from edsl.Base import Base
from edsl.utilities.remove_edsl_version import remove_edsl_version
from edsl.exceptions.agents import AgentListError
from edsl.utilities.is_notebook import is_notebook
from edsl.results.ResultsExportMixin import ResultsExportMixin
import logging

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from edsl.scenarios.ScenarioList import ScenarioList
    from edsl.agents.Agent import Agent
    from pandas import DataFrame


def is_iterable(obj):
    return isinstance(obj, Iterable)


class EmptyAgentList:
    def __repr__(self):
        return "Empty AgentList"


# ResultsExportMixin,
class AgentList(UserList, Base):
    """A list of Agents."""

    __documentation__ = (
        "https://docs.expectedparrot.com/en/latest/agents.html#agentlist-class"
    )

    def __init__(self, data: Optional[list["Agent"]] = None):
        """Initialize a new AgentList.

        :param data: A list of Agents.
        """
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def shuffle(self, seed: Optional[str] = None) -> AgentList:
        """Shuffle the AgentList.

        :param seed: The seed for the random number generator.
        """
        import random

        if seed is not None:
            random.seed(seed)
        random.shuffle(self.data)
        return self

    def sample(self, n: int, seed: Optional[str] = None) -> AgentList:
        """Return a random sample of agents.

        :param n: The number of agents to sample.
        :param seed: The seed for the random number generator.
        """
        import random

        if seed:
            random.seed(seed)
        return AgentList(random.sample(self.data, n))

    def to_pandas(self) -> "DataFrame":
        """Return a pandas DataFrame.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> al.to_pandas()
           age   hair  height
        0   22  brown     5.5
        1   22  brown     5.5
        """
        return self.to_scenario_list().to_pandas()

    def tally(
        self, *fields: Optional[str], top_n: Optional[int] = None, output="Dataset"
    ) -> Union[dict, "Dataset"]:
        """Tally the values of a field or perform a cross-tab of multiple fields.

        :param fields: The field(s) to tally, multiple fields for cross-tabulation.

        >>> al = AgentList.example()
        >>> al.tally('age')
        Dataset([{'age': [22]}, {'count': [2]}])
        """
        return self.to_scenario_list().tally(*fields, top_n=top_n, output=output)

    def duplicate(self):
        """Duplicate the AgentList.

        >>> al = AgentList.example()
        >>> al2 = al.duplicate()
        >>> al2 == al
        True
        >>> id(al2) == id(al)
        False
        """
        return AgentList([a.duplicate() for a in self.data])

    def rename(self, old_name, new_name) -> AgentList:
        """Rename a trait in the AgentList.

        :param old_name: The old name of the trait.
        :param new_name: The new name of the trait.
        :param inplace: Whether to rename the trait in place.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al2 = al.rename('a', 'c')
        >>> assert al2 == AgentList([Agent(traits = {'c': 1, 'b': 1}), Agent(traits = {'c': 1, 'b': 2})])
        >>> assert al != al2
        """
        newagents = []
        for agent in self:
            newagents.append(agent.rename(old_name, new_name))
        return AgentList(newagents)

    def select(self, *traits) -> AgentList:
        """Selects agents with only the references traits.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.select('a')
        AgentList([Agent(traits = {'a': 1}), Agent(traits = {'a': 1})])

        """

        if len(traits) == 1:
            traits_to_select = [list(traits)[0]]
        else:
            traits_to_select = list(traits)

        return AgentList([agent.select(*traits_to_select) for agent in self.data])

    def filter(self, expression: str) -> AgentList:
        """
        Filter a list of agents based on an expression.

        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent(traits = {'a': 1, 'b': 1}), Agent(traits = {'a': 1, 'b': 2})])
        >>> al.filter("b == 2")
        AgentList([Agent(traits = {'a': 1, 'b': 2})])
        """

        def create_evaluator(agent: "Agent"):
            """Create an evaluator for the given result.
            The 'combined_dict' is a mapping of all values for that Result object.
            """
            return EvalWithCompoundTypes(names=agent.traits)

            # iterates through all the results and evaluates the expression

        try:
            new_data = [
                agent for agent in self.data if create_evaluator(agent).eval(expression)
            ]
        except NameNotDefined as e:
            e = AgentListError(f"'{expression}' is not a valid expression.")
            if is_notebook():
                print(e, file=sys.stderr)
            else:
                raise e

            return EmptyAgentList()

        if len(new_data) == 0:
            return EmptyAgentList()

        return AgentList(new_data)

    @property
    def all_traits(self) -> list[str]:
        """Return all traits in the AgentList.
        >>> from edsl.agents.Agent import Agent
        >>> agent_1 = Agent(traits = {'age': 22})
        >>> agent_2 = Agent(traits = {'hair': 'brown'})
        >>> al = AgentList([agent_1, agent_2])
        >>> al.all_traits
        ['age', 'hair']
        """
        d = {}
        for agent in self:
            d.update(agent.traits)
        return list(d.keys())

    @classmethod
    def from_csv(cls, file_path: str, name_field: Optional[str] = None):
        """Load AgentList from a CSV file.

        >>> import csv
        >>> import os
        >>> with open('/tmp/agents.csv', 'w') as f:
        ...     writer = csv.writer(f)
        ...     _ = writer.writerow(['age', 'hair', 'height'])
        ...     _ = writer.writerow([22, 'brown', 5.5])
        >>> al = AgentList.from_csv('/tmp/agents.csv')
        >>> al
        AgentList([Agent(traits = {'age': '22', 'hair': 'brown', 'height': '5.5'})])
        >>> al = AgentList.from_csv('/tmp/agents.csv', name_field='hair')
        >>> al
        AgentList([Agent(name = \"""brown\""", traits = {'age': '22', 'height': '5.5'})])
        >>> os.remove('/tmp/agents.csv')

        :param file_path: The path to the CSV file.
        :param name_field: The name of the field to use as the agent name.
        """
        from edsl.agents.Agent import Agent

        agent_list = []
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            for row in reader:
                if "name" in row:
                    import warnings

                    warnings.warn("Using 'name' field in the CSV for the Agent name")
                    name_field = "name"
                if name_field is not None:
                    agent_name = row.pop(name_field)
                    agent_list.append(Agent(traits=row, name=agent_name))
                else:
                    agent_list.append(Agent(traits=row))
        return cls(agent_list)

    def translate_traits(self, codebook: dict[str, str]):
        """Translate traits to a new codebook.

        :param codebook: The new codebook.

        >>> al = AgentList.example()
        >>> codebook = {'hair': {'brown':'Secret word for green'}}
        >>> al.translate_traits(codebook)
        AgentList([Agent(traits = {'age': 22, 'hair': 'Secret word for green', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'Secret word for green', 'height': 5.5})])
        """
        new_agents = []
        for agent in self.data:
            new_agents.append(agent.translate_traits(codebook))
        return AgentList(new_agents)

    def remove_trait(self, trait: str):
        """Remove traits from the AgentList.

        :param traits: The traits to remove.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent({'age': 22, 'hair': 'brown', 'height': 5.5}), Agent({'age': 22, 'hair': 'brown', 'height': 5.5})])
        >>> al.remove_trait('age')
        AgentList([Agent(traits = {'hair': 'brown', 'height': 5.5}), Agent(traits = {'hair': 'brown', 'height': 5.5})])
        """
        agents = []
        new_al = self.duplicate()
        for agent in new_al.data:
            agents.append(agent.remove_trait(trait))
        return AgentList(agents)

    def add_trait(self, trait: str, values: List[Any]) -> AgentList:
        """Adds a new trait to every agent, with values taken from values.

        :param trait: The name of the trait.
        :param values: The valeues(s) of the trait. If a single value is passed, it is used for all agents.

        >>> al = AgentList.example()
        >>> new_al = al.add_trait('new_trait', 1)
        >>> new_al.select('new_trait').to_scenario_list().to_list()
        [1, 1]
        >>> al.add_trait('new_trait', [1, 2, 3])
        Traceback (most recent call last):
        ...
        edsl.exceptions.agents.AgentListError: The passed values have to be the same length as the agent list.
        ...
        """
        if not is_iterable(values):
            new_agents = []
            value = values
            for agent in self.data:
                new_agents.append(agent.add_trait(trait, value))
            return AgentList(new_agents)

        if len(values) != len(self):
            e = AgentListError(
                "The passed values have to be the same length as the agent list."
            )
            if is_notebook():
                print(e, file=sys.stderr)
            else:
                raise e
        new_agents = []
        for agent, value in zip(self.data, values):
            new_agents.append(agent.add_trait(trait, value))
        return AgentList(new_agents)

    @staticmethod
    def get_codebook(file_path: str):
        """Return the codebook for a CSV file.

        :param file_path: The path to the CSV file.
        """
        with open(file_path, "r") as f:
            reader = csv.DictReader(f)
            return {field: None for field in reader.fieldnames}

    def __hash__(self) -> int:
        """Return the hash of the AgentList.

        >>> al = AgentList.example()
        >>> hash(al)
        1681154913465662422
        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False, sorted=True))

    def to_dict(self, sorted=False, add_edsl_version=True):
        """Serialize the AgentList to a dictionary.

        >>> AgentList.example().to_dict(add_edsl_version=False)
        {'agent_list': [{'traits': {'age': 22, 'hair': 'brown', 'height': 5.5}}, {'traits': {'age': 22, 'hair': 'brown', 'height': 5.5}}]}

        """
        if sorted:
            data = self.data[:]
            data.sort(key=lambda x: hash(x))
        else:
            data = self.data

        d = {
            "agent_list": [
                agent.to_dict(add_edsl_version=add_edsl_version) for agent in data
            ]
        }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "AgentList"

        return d

    def __eq__(self, other: AgentList) -> bool:
        return self.to_dict(sorted=True, add_edsl_version=False) == other.to_dict(
            sorted=True, add_edsl_version=False
        )

    def __repr__(self):
        return f"AgentList({self.data})"

    def _summary(self):
        return {
            "agents": len(self),
        }

    def set_codebook(self, codebook: dict[str, str]) -> AgentList:
        """Set the codebook for the AgentList.

        >>> from edsl.agents.Agent import Agent
        >>> a = Agent(traits = {'hair': 'brown'})
        >>> al = AgentList([a, a])
        >>> _ = al.set_codebook({'hair': "Color of hair on driver's license"})
        >>> al[0].codebook
        {'hair': "Color of hair on driver's license"}


        :param codebook: The codebook.
        """
        for agent in self.data:
            agent.codebook = codebook

        return self

    def to_csv(self, file_path: str):
        """Save the AgentList to a CSV file.

        :param file_path: The path to the CSV file.
        """
        self.to_scenario_list().to_csv(file_path)

    def to_list(self, include_agent_name=False) -> list[tuple]:
        """Return a list of tuples."""
        return self.to_scenario_list(include_agent_name).to_list()

    def to_scenario_list(
        self, include_agent_name: bool = False, include_instruction: bool = False
    ) -> ScenarioList:
        """Converts the agent to a scenario list."""
        from edsl.scenarios.ScenarioList import ScenarioList
        from edsl.scenarios.Scenario import Scenario

        # raise NotImplementedError("This method is not implemented yet.")

        scenario_list = ScenarioList()
        for agent in self.data:
            d = agent.traits
            if include_agent_name:
                d["agent_name"] = agent.name
            if include_instruction:
                d["instruction"] = agent.instruction
            scenario_list.append(Scenario(d))
        return scenario_list

        # if include_agent_name:
        #     return ScenarioList(
        #         [
        #             Scenario(agent.traits | {"agent_name": agent.name} | })
        #             for agent in self.data
        #         ]
        #     )
        # return ScenarioList([Scenario(agent.traits) for agent in self.data])

    def table(
        self,
        *fields,
        tablefmt: Optional[str] = None,
        pretty_labels: Optional[dict] = None,
    ) -> Table:
        if len(self) == 0:
            e = AgentListError("Cannot create a table from an empty AgentList.")
            if is_notebook():
                print(e, file=sys.stderr)
                return None
            else:
                raise e
        return (
            self.to_scenario_list()
            .to_dataset()
            .table(*fields, tablefmt=tablefmt, pretty_labels=pretty_labels)
        )

    def to_dataset(self, traits_only: bool = True):
        """
        Convert the AgentList to a Dataset.

        >>> from edsl.agents.AgentList import AgentList
        >>> al = AgentList.example()
        >>> al.to_dataset()
        Dataset([{'age': [22, 22]}, {'hair': ['brown', 'brown']}, {'height': [5.5, 5.5]}])
        >>> al.to_dataset(traits_only = False)
        Dataset([{'age': [22, 22]}, {'hair': ['brown', 'brown']}, {'height': [5.5, 5.5]}, {'agent_parameters': [{'instruction': 'You are answering questions as if you were a human. Do not break character.', 'agent_name': None}, {'instruction': 'You are answering questions as if you were a human. Do not break character.', 'agent_name': None}]}])
        """
        from edsl.results.Dataset import Dataset
        from collections import defaultdict

        agent_trait_keys = []
        for agent in self:
            agent_keys = list(agent.traits.keys())
            for key in agent_keys:
                if key not in agent_trait_keys:
                    agent_trait_keys.append(key)

        data = defaultdict(list)
        for agent in self:
            for trait_key in agent_trait_keys:
                data[trait_key].append(agent.traits.get(trait_key, None))
            if not traits_only:
                data["agent_parameters"].append(
                    {"instruction": agent.instruction, "agent_name": agent.name}
                )
        return Dataset([{key: entry} for key, entry in data.items()])

    def tree(self, node_order: Optional[List[str]] = None):
        return self.to_scenario_list().tree(node_order)

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserialize the dictionary back to an AgentList object.

        :param: data: A dictionary representing an AgentList.
        >>> from edsl.agents.Agent import Agent
        >>> al = AgentList([Agent.example(), Agent.example()])
        >>> al2 = AgentList.from_dict(al.to_dict())
        >>> al2 == al
        True
        """
        from edsl.agents.Agent import Agent

        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls, randomize: bool = False) -> AgentList:
        """
        Returns an example AgentList instance.

        :param randomize: If True, uses Agent's randomize method.
        """
        from edsl.agents.Agent import Agent

        return cls([Agent.example(randomize), Agent.example(randomize)])

    @classmethod
    def from_list(self, trait_name: str, values: List[Any]):
        """Create an AgentList from a list of values.

        :param trait_name: The name of the trait.
        :param values: A list of values.

        >>> AgentList.from_list('age', [22, 23])
        AgentList([Agent(traits = {'age': 22}), Agent(traits = {'age': 23})])
        """
        from edsl.agents.Agent import Agent

        return AgentList([Agent({trait_name: value}) for value in values])

    def __mul__(self, other: AgentList) -> AgentList:
        """Takes the cross product of two AgentLists."""
        from itertools import product

        new_sl = []
        for s1, s2 in list(product(self, other)):
            new_sl.append(s1 + s2)
        return AgentList(new_sl)

    def code(self, string=True) -> Union[str, list[str]]:
        """Return code to construct an AgentList.

        >>> al = AgentList.example()
        >>> print(al.code())
        from edsl.agents.Agent import Agent
        from edsl.agents.AgentList import AgentList
        agent_list = AgentList([Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5}), Agent(traits = {'age': 22, 'hair': 'brown', 'height': 5.5})])
        """
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        if string:
            return "\n".join(lines)
        return lines


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""A Scenario is a dictionary with a key/value to parameterize a question."""

from __future__ import annotations
import copy
import os
import json
from collections import UserDict
from typing import Union, List, Optional, TYPE_CHECKING, Collection
from uuid import uuid4

from edsl.Base import Base
from edsl.scenarios.ScenarioHtmlMixin import ScenarioHtmlMixin
from edsl.utilities.remove_edsl_version import remove_edsl_version
from edsl.exceptions.scenarios import ScenarioError

if TYPE_CHECKING:
    from edsl.scenarios.ScenarioList import ScenarioList
    from edsl.results.Dataset import Dataset


class DisplayJSON:
    """Display a dictionary as JSON."""

    def __init__(self, input_dict: dict):
        self.text = json.dumps(input_dict, indent=4)

    def __repr__(self):
        return self.text


class DisplayYAML:
    """Display a dictionary as YAML."""

    def __init__(self, input_dict: dict):
        import yaml

        self.text = yaml.dump(input_dict)

    def __repr__(self):
        return self.text


class Scenario(Base, UserDict, ScenarioHtmlMixin):
    """A Scenario is a dictionary of keys/values that can be used to parameterize questions."""

    __documentation__ = "https://docs.expectedparrot.com/en/latest/scenarios.html"

    def __init__(self, data: Optional[dict] = None, name: Optional[str] = None):
        """Initialize a new Scenario.

        :param data: A dictionary of keys/values for parameterizing questions.
        :param name: The name of the scenario.
        """
        if not isinstance(data, dict) and data is not None:
            try:
                data = dict(data)
            except Exception as e:
                raise ScenarioError(
                    f"You must pass in a dictionary to initialize a Scenario. You passed in {data}",
                    "Exception message:" + str(e),
                )

        super().__init__()
        self.data = data if data is not None else {}
        self.name = name

    def replicate(self, n: int) -> "ScenarioList":
        """Replicate a scenario n times to return a ScenarioList.

        :param n: The number of times to replicate the scenario.

        Example:
        >>> s = Scenario({"food": "wood chips"})
        >>> s.replicate(2)
        ScenarioList([Scenario({'food': 'wood chips'}), Scenario({'food': 'wood chips'})])
        """
        from edsl.scenarios.ScenarioList import ScenarioList

        return ScenarioList([copy.deepcopy(self) for _ in range(n)])

    @property
    def has_jinja_braces(self) -> bool:
        """Return whether the scenario has jinja braces. This matters for rendering.

        >>> s = Scenario({"food": "I love {{wood chips}}"})
        >>> s.has_jinja_braces
        True
        """
        for _, value in self.items():
            if isinstance(value, str):
                if "{{" in value and "}}" in value:
                    return True
        return False

    def _convert_jinja_braces(
        self, replacement_left: str = "<<", replacement_right: str = ">>"
    ) -> Scenario:
        """Convert Jinja braces to some other character.

        >>> s = Scenario({"food": "I love {{wood chips}}"})
        >>> s._convert_jinja_braces()
        Scenario({'food': 'I love <<wood chips>>'})

        """
        new_scenario = Scenario()
        for key, value in self.items():
            if isinstance(value, str):
                new_scenario[key] = value.replace("{{", replacement_left).replace(
                    "}}", replacement_right
                )
            else:
                new_scenario[key] = value
        return new_scenario

    def __add__(self, other_scenario: Scenario) -> Scenario:
        """Combine two scenarios by taking the union of their keys

        If the other scenario is None, then just return self.

        :param other_scenario: The other scenario to combine with.

        Example:

        >>> s1 = Scenario({"price": 100, "quantity": 2})
        >>> s2 = Scenario({"color": "red"})
        >>> s1 + s2
        Scenario({'price': 100, 'quantity': 2, 'color': 'red'})
        >>> (s1 + s2).__class__.__name__
        'Scenario'
        """
        if other_scenario is None:
            return self
        else:
            data1 = copy.deepcopy(self.data)
            data2 = copy.deepcopy(other_scenario.data)
            s = Scenario(data1 | data2)
            return s

    def rename(
        self,
        old_name_or_replacement_dict: Union[str, dict[str, str]],
        new_name: Optional[str] = None,
    ) -> Scenario:
        """Rename the keys of a scenario.

        :param old_name_or_replacement_dict: A dictionary of old keys to new keys *OR* a string of the old key.
        :param new_name: The new name of the key.

        Example:

        >>> s = Scenario({"food": "wood chips"})
        >>> s.rename({"food": "food_preference"})
        Scenario({'food_preference': 'wood chips'})

        >>> s = Scenario({"food": "wood chips"})
        >>> s.rename("food", "snack")
        Scenario({'snack': 'wood chips'})
        """
        if isinstance(old_name_or_replacement_dict, str) and new_name is not None:
            replacement_dict = {old_name_or_replacement_dict: new_name}
        else:
            replacement_dict = old_name_or_replacement_dict

        new_scenario = Scenario()
        for key, value in self.items():
            if key in replacement_dict:
                new_scenario[replacement_dict[key]] = value
            else:
                new_scenario[key] = value
        return new_scenario

    def new_column_names(self, new_names: List[str]) -> Scenario:
        """Rename the keys of a scenario.

        >>> s = Scenario({"food": "wood chips"})
        >>> s.new_column_names(["food_preference"])
        Scenario({'food_preference': 'wood chips'})
        """
        try:
            assert len(new_names) == len(self.keys())
        except AssertionError:
            print("The number of new names must match the number of keys.")

        new_scenario = Scenario()
        for new_names, value in zip(new_names, self.values()):
            new_scenario[new_names] = value
        return new_scenario

    def table(self, tablefmt: str = "grid") -> str:
        """Display a scenario as a table."""
        return self.to_dataset().table(tablefmt=tablefmt)

    def json(self):
        return DisplayJSON(self.to_dict(add_edsl_version=False))

    def yaml(self):
        import yaml

        return DisplayYAML(self.to_dict(add_edsl_version=False))

    def to_dict(self, add_edsl_version: bool = True) -> dict:
        """Convert a scenario to a dictionary.

        Example:

        >>> s = Scenario({"food": "wood chips"})
        >>> s.to_dict()
        {'food': 'wood chips', 'edsl_version': '...', 'edsl_class_name': 'Scenario'}

        >>> s.to_dict(add_edsl_version = False)
        {'food': 'wood chips'}

        """
        from edsl.scenarios.FileStore import FileStore

        d = self.data.copy()
        for key, value in d.items():
            if isinstance(value, FileStore):
                d[key] = value.to_dict(add_edsl_version=add_edsl_version)
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "Scenario"

        return d

    def __hash__(self) -> int:
        """Return a hash of the scenario.

        Example:

        >>> s = Scenario({"food": "wood chips"})
        >>> hash(s)
        1153210385458344214
        """
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False))

    def __repr__(self):
        return "Scenario(" + repr(self.data) + ")"

    def to_dataset(self) -> "Dataset":
        """Convert a scenario to a dataset.

        >>> s = Scenario({"food": "wood chips"})
        >>> s.to_dataset()
        Dataset([{'key': ['food']}, {'value': ['wood chips']}])
        """
        from edsl.results.Dataset import Dataset

        keys = list(self.keys())
        values = list(self.values())
        return Dataset([{"key": keys}, {"value": values}])

    def select(self, list_of_keys: Collection[str]) -> "Scenario":
        """Select a subset of keys from a scenario.

        :param list_of_keys: The keys to select.

        Example:

        >>> s = Scenario({"food": "wood chips", "drink": "water"})
        >>> s.select(["food"])
        Scenario({'food': 'wood chips'})
        """
        new_scenario = Scenario()
        for key in list_of_keys:
            new_scenario[key] = self[key]
        return new_scenario

    def drop(self, list_of_keys: Collection[str]) -> "Scenario":
        """Drop a subset of keys from a scenario.

        :param list_of_keys: The keys to drop.

        Example:

        >>> s = Scenario({"food": "wood chips", "drink": "water"})
        >>> s.drop(["food"])
        Scenario({'drink': 'water'})
        """
        new_scenario = Scenario()
        for key in self.keys():
            if key not in list_of_keys:
                new_scenario[key] = self[key]
        return new_scenario

    def keep(self, list_of_keys: List[str]) -> "Scenario":
        """Keep a subset of keys from a scenario.

        :param list_of_keys: The keys to keep.

        Example:

        >>> s = Scenario({"food": "wood chips", "drink": "water"})
        >>> s.keep(["food"])
        Scenario({'food': 'wood chips'})
        """

        return self.select(list_of_keys)

    @classmethod
    def from_url(cls, url: str, field_name: Optional[str] = "text") -> "Scenario":
        """Creates a scenario from a URL.

        :param url: The URL to create the scenario from.
        :param field_name: The field name to use for the text.

        """
        import requests

        text = requests.get(url).text
        return cls({"url": url, field_name: text})

    @classmethod
    def from_file(cls, file_path: str, field_name: str) -> "Scenario":
        """Creates a scenario from a file.

        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile(suffix=".txt", mode="w") as f:
        ...     _ = f.write("This is a test.")
        ...     _ = f.flush()
        ...     s = Scenario.from_file(f.name, "file")
        >>> s
        Scenario({'file': FileStore(path='...', ...)})

        """
        from edsl.scenarios.FileStore import FileStore

        fs = FileStore(file_path)
        return cls({field_name: fs})

    @classmethod
    def from_image(
        cls, image_path: str, image_name: Optional[str] = None
    ) -> "Scenario":
        """
        Creates a scenario with a base64 encoding of an image.

        Args:
            image_path (str): Path to the image file.

        Returns:
            Scenario: A new Scenario instance with image information.

        """
        if not os.path.exists(image_path):
            raise FileNotFoundError(f"Image file not found: {image_path}")

        if image_name is None:
            image_name = os.path.basename(image_path).split(".")[0]

        return cls.from_file(image_path, image_name)

    @classmethod
    def from_pdf(cls, pdf_path: str):
        from edsl.scenarios.PdfExtractor import PdfExtractor

        return PdfExtractor(pdf_path, cls).get_object()

    @classmethod
    def from_docx(cls, docx_path: str) -> "Scenario":
        """Creates a scenario from the text of a docx file.

        :param docx_path: The path to the docx file.

        Example:

        >>> from docx import Document
        >>> doc = Document()
        >>> _ = doc.add_heading("EDSL Survey")
        >>> _ = doc.add_paragraph("This is a test.")
        >>> doc.save("test.docx")
        >>> s = Scenario.from_docx("test.docx")
        >>> s
        Scenario({'file_path': 'test.docx', 'text': 'EDSL Survey\\nThis is a test.'})
        >>> import os; os.remove("test.docx")
        """
        from edsl.scenarios.DocxScenario import DocxScenario

        return Scenario(DocxScenario(docx_path).get_scenario_dict())

    def chunk(
        self,
        field,
        num_words: Optional[int] = None,
        num_lines: Optional[int] = None,
        include_original=False,
        hash_original=False,
    ) -> "ScenarioList":
        """Split a field into chunks of a given size.

        :param field: The field to split.
        :param num_words: The number of words in each chunk.
        :param num_lines: The number of lines in each chunk.
        :param include_original: Whether to include the original field in the new scenarios.
        :param hash_original: Whether to hash the original field in the new scenarios.

        If you specify `include_original=True`, the original field will be included in the new scenarios with an "_original" suffix.

        Either `num_words` or `num_lines` must be specified, but not both.

        The `hash_original` parameter is useful if you do not want to store the original text, but still want a unique identifier for it.

        Example:

        >>> s = Scenario({"text": "This is a test.\\nThis is a test.\\n\\nThis is a test."})
        >>> s.chunk("text", num_lines = 1)
        ScenarioList([Scenario({'text': 'This is a test.', 'text_chunk': 0}), Scenario({'text': 'This is a test.', 'text_chunk': 1}), Scenario({'text': '', 'text_chunk': 2}), Scenario({'text': 'This is a test.', 'text_chunk': 3})])

        >>> s.chunk("text", num_words = 2)
        ScenarioList([Scenario({'text': 'This is', 'text_chunk': 0}), Scenario({'text': 'a test.', 'text_chunk': 1}), Scenario({'text': 'This is', 'text_chunk': 2}), Scenario({'text': 'a test.', 'text_chunk': 3}), Scenario({'text': 'This is', 'text_chunk': 4}), Scenario({'text': 'a test.', 'text_chunk': 5})])

        >>> s = Scenario({"text": "Hello World"})
        >>> s.chunk("text", num_words = 1, include_original = True)
        ScenarioList([Scenario({'text': 'Hello', 'text_chunk': 0, 'text_original': 'Hello World'}), Scenario({'text': 'World', 'text_chunk': 1, 'text_original': 'Hello World'})])

        >>> s = Scenario({"text": "Hello World"})
        >>> s.chunk("text", num_words = 1, include_original = True, hash_original = True)
        ScenarioList([Scenario({'text': 'Hello', 'text_chunk': 0, 'text_original': 'b10a8db164e0754105b7a99be72e3fe5'}), Scenario({'text': 'World', 'text_chunk': 1, 'text_original': 'b10a8db164e0754105b7a99be72e3fe5'})])

        >>> s.chunk("text")
        Traceback (most recent call last):
        ...
        ValueError: You must specify either num_words or num_lines.

        >>> s.chunk("text", num_words = 1, num_lines = 1)
        Traceback (most recent call last):
        ...
        ValueError: You must specify either num_words or num_lines, but not both.
        """
        from edsl.scenarios.DocumentChunker import DocumentChunker

        return DocumentChunker(self).chunk(
            field, num_words, num_lines, include_original, hash_original
        )

    @classmethod
    @remove_edsl_version
    def from_dict(cls, d: dict) -> "Scenario":
        """Convert a dictionary to a scenario.

        Example:

        >>> Scenario.from_dict({"food": "wood chips"})
        Scenario({'food': 'wood chips'})
        """
        from edsl.scenarios.FileStore import FileStore

        for key, value in d.items():
            # TODO: we should check this better if its a FileStore + add remote security check against path traversal
            if (
                isinstance(value, dict) and "base64_string" in value and "path" in value
            ) or isinstance(value, FileStore):
                d[key] = FileStore.from_dict(value)
        return cls(d)

    def _table(self) -> tuple[dict, list]:
        """Prepare generic table data.
        >>> s = Scenario({"food": "wood chips"})
        >>> s._table()
        ([{'Attribute': 'data', 'Value': "{'food': 'wood chips'}"}, {'Attribute': 'name', 'Value': 'None'}], ['Attribute', 'Value'])
        """
        table_data = []
        for attr_name, attr_value in self.__dict__.items():
            table_data.append({"Attribute": attr_name, "Value": repr(attr_value)})
        column_names = ["Attribute", "Value"]
        return table_data, column_names

    @classmethod
    def example(cls, randomize: bool = False) -> Scenario:
        """
        Returns an example Scenario instance.

        :param randomize: If True, adds a random string to the value of the example key.
        """
        addition = "" if not randomize else str(uuid4())
        return cls(
            {
                "persona": f"A reseacher studying whether LLMs can be used to generate surveys.{addition}",
            }
        )

    def code(self) -> List[str]:
        """Return the code for the scenario."""
        lines = []
        lines.append("from edsl.scenario import Scenario")
        lines.append(f"s = Scenario({self.data})")
        # return f"Scenario({self.data})"
        return lines


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from typing import Callable, Union, List
from collections import UserDict

from edsl.jobs.tokens.TokenUsage import TokenUsage
from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage


class TaskCreators(UserDict):
    """A dictionary of task creators. A task is one question being answered.

    This is used to track the status of the tasks within an interview.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def token_usage(self) -> InterviewTokenUsage:
        """Determines how many tokens were used for the interview.

        This is iterates through all tasks that make up an interview.
        For each task, it determines how many tokens were used and whether they were cached or new.
        It then sums the total number of cached and new tokens used for the interview.

        """
        cached_tokens = TokenUsage(from_cache=True)
        new_tokens = TokenUsage(from_cache=False)
        for task_creator in self.values():
            token_usage = task_creator.token_usage()
            cached_tokens += token_usage["cached_tokens"]
            new_tokens += token_usage["new_tokens"]
        return InterviewTokenUsage(
            new_token_usage=new_tokens, cached_token_usage=cached_tokens
        )

    def print(self) -> None:
        from rich import print

        print({task.get_name(): task.task_status for task in self.values()})

    @property
    def interview_status(self) -> InterviewStatusDictionary:
        """Returns a dictionary, InterviewStatusDictionary, mapping task status codes to counts of tasks in that state.

        >>> t = TaskCreators()
        >>> t.interview_status
        InterviewStatusDictionary({<TaskStatus.NOT_STARTED: 1>: 0, <TaskStatus.WAITING_FOR_DEPENDENCIES: 2>: 0, <TaskStatus.CANCELLED: 3>: 0, <TaskStatus.PARENT_FAILED: 4>: 0, <TaskStatus.WAITING_FOR_REQUEST_CAPACITY: 5>: 0, <TaskStatus.WAITING_FOR_TOKEN_CAPACITY: 6>: 0, <TaskStatus.API_CALL_IN_PROGRESS: 7>: 0, <TaskStatus.SUCCESS: 8>: 0, <TaskStatus.FAILED: 9>: 0, 'number_from_cache': 0})
        """
        status_dict = InterviewStatusDictionary()
        for task_creator in self.values():
            status_dict[task_creator.task_status] += 1
            status_dict["number_from_cache"] += task_creator.from_cache
        return status_dict

    def status_logs(self):
        """Returns a list of status logs for each task."""
        return [task_creator.status_log for task_creator in self.values()]


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
from typing import Optional
from edsl.questions.descriptors import QuestionOptionsDescriptor
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

from edsl.questions.decorators import inject_exception


class QuestionYesNo(QuestionMultipleChoice):
    """This question prompts the agent to respond with 'Yes' or 'No'."""

    question_type = "yes_no"
    question_options = QuestionOptionsDescriptor(num_choices=2)

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str] = ["No", "Yes"],
        answering_instructions: Optional[str] = None,
        question_presentation: Optional[str] = None,
        include_comment: Optional[bool] = True,
    ):
        """Instantiate a new QuestionYesNo.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionYesNo.default_instructions`.
        """
        super().__init__(
            question_name=question_name,
            question_text=question_text,
            question_options=question_options,
            use_code=False,
            answering_instructions=answering_instructions,
            question_presentation=question_presentation,
            include_comment=include_comment,
        )
        self.question_options = question_options

    ################
    # Helpful
    ################
    @classmethod
    @inject_exception
    def example(cls, include_comment: bool = True) -> QuestionYesNo:
        """Return an example of a yes/no question."""
        return cls(
            question_name="is_it_equal",
            question_text="Is 5 + 5 equal to 11?",
            include_comment=include_comment,
        )


def main():
    """Create an example of a yes/no question and demonstrate its functionality."""
    from edsl.questions.derived.QuestionYesNo import QuestionYesNo

    q = QuestionYesNo.example()
    q.question_text
    q.question_options
    q.question_name
    # validate an answer
    q._validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer(0, {})
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from abc import ABC, abstractmethod
import json
from typing import Dict, List, Any, TypeVar, Generator, Dict, Callable
from dataclasses import dataclass, field, KW_ONLY, fields, asdict
import textwrap


class ExceptionPipesDoNotFit(Exception):
    pass


class StageProcessingClosure:
    def __init__(self, stage_func: Callable, reduction_func=lambda x: x):
        self.data = []
        self.stage_func = stage_func
        # reduction function is applied to self.data when complete
        # it might just return the list, or it might do something more complicated such as
        # reduce the list to a dictionary
        self.reduction_func = reduction_func

    def func(self, obj: "FlowDataBase") -> None:
        "Function to apply to each stage"
        self.data.append(self.stage_func(obj))

    def __call__(self):
        return self.reduction_func(self.data)


@dataclass
class FlowDataBase:
    """Base class for dataclasses that are passed between stages."""

    _: KW_ONLY
    # previous_stage: Dict = field(default_factory=dict)
    previous_stage: Any = None
    sent_to_stage_name: str = field(default_factory=str)
    came_from_stage_name: str = field(default_factory=str)

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def __getitem__(self, key):
        """Allows dictionary-style getting."""
        return getattr(self, key)

    def __setitem__(self, key, value):
        """Allows dictionary-style setting."""
        return setattr(self, key, value)

    def current_values(self):
        """Returns a dictionary of the current values of the dataclass"""
        to_exclude = ["sent_to_stage_name", "came_from_stage_name", "previous_stage"]
        d = asdict(self)
        [d.pop(key) for key in to_exclude]
        return d

    def stage_input_output(self):
        return {
            "came_from": self.came_from_stage_name,
            "sent_to": self.sent_to_stage_name,
        }

    def _align_values_with_padding(
        self, stages
    ) -> Generator[Dict[str, str], None, None]:
        "Pads out the the names of the stages so they are aligned when printing"

        def longest_value(stage):
            return max([len(v) for v in stage.values()])

        max_length = max([longest_value(stage) for stage in stages])
        for stage in stages:
            new_stage = {k: v.ljust(max_length) for k, v in stage.items()}
            yield new_stage

    def _reduce(self, stage_processor: StageProcessingClosure) -> Dict[str, dict]:
        """Applies some function defined in stage_processor to each stage in the chain, working from back to front

        The stage_processor will record the results of the function applied to each stage in
        an instance of the StageProcessingClosure class.
        The results can be accessed by calling the StageProcessingClosure instance.
        This somewhat convoluted approach is necessary because the stages are connected in a chain and
        we want a way to access the results of the function applied to each stage in the chain without
        writing the while-loop over and over again.
        """
        stage_processor.func(self)
        current_pipe = self
        while True:
            if current_pipe.previous_stage is None:
                break
            else:
                current_pipe = current_pipe.previous_stage
                stage_processor.func(
                    current_pipe
                )  # the result is getting stored in stage_processor.data

    def combined_results(self) -> Dict[str, dict]:
        stage_processor = StageProcessingClosure(
            stage_func=lambda obj: obj.current_values(),
            reduction_func=lambda x: {k: v for d in x for k, v in d.items()},
        )
        self._reduce(stage_processor)
        return stage_processor()

    def flow_history(self):
        stage_processor = StageProcessingClosure(
            stage_func=lambda obj: obj.stage_input_output()
        )
        self._reduce(stage_processor)
        return stage_processor()

    def visualize_flow(self) -> str:
        """Visualize the flow of data through the chain"""
        stages = self.flow_history()
        new_stages = list(self._align_values_with_padding(stages))
        new_stages.reverse()
        return tuple(new_stages)


class StageBase(ABC):
    input: FlowDataBase = NotImplemented
    output: FlowDataBase = NotImplemented

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

        if hasattr(self, "next_stage"):
            self._validate_connection(self.next_stage)
        else:
            self.next_stage = None

    @classmethod
    def function_parameters(self):
        return fields(self.input)

    @classmethod
    def func(cls, **kwargs):
        "This provides a shortcut for running a stage by passing keyword arguments to the input function."
        input_data = cls.input(**kwargs)
        return cls().process(input_data)

    @abstractmethod
    def handle_data(self, data):
        "This implements how the stage actually handles the passed in data"
        raise NotImplementedError

    def _validate_connection(self, stage):
        "Checks that the outputs of the first stage match the inputs of the second stage"
        if not self.output == stage.input:
            raise ExceptionPipesDoNotFit(
                textwrap.dedent(
                    f"""\
        Stage \"{self.__class__.__name__}\" cannot be connected to stage \"{stage.__class__.__name__}\".
        The outputs of the first stage {self.output} do not match the inputs of the second stage, {stage.input}."""
                )
            )

    def __init_subclass__(cls, **kwargs):
        "Checks that the subclass has the required class variables of input & output"
        super().__init_subclass__(**kwargs)
        if cls.input is NotImplemented:
            raise NotImplementedError(
                f"Class {cls.__name__} lacks required class variable 'inputs'"
            )
        if cls.output is NotImplemented:
            raise NotImplementedError(
                f"Class {cls.__name__} lacks required class variable 'outputs'"
            )

    def process(self, data):
        print(f"Running stage: {self.__class__.__name__}")
        data.sent_to_stage_name = self.__class__.__name__
        processed_data = self.handle_data(data)
        processed_data.came_from_stage_name = self.__class__.__name__
        processed_data.previous_stage = data
        if self.next_stage:
            return self.next_stage.process(processed_data)
        else:
            return processed_data


if __name__ == "__main__":
    pass
    # try:

    #     class StageMissing(StageBase):
    #         def handle_data(self, data):
    #             return data

    # except NotImplementedError as e:
    #     print(e)
    # else:
    #     raise Exception("Should have raised NotImplementedError")

    # try:

    #     class StageMissingInput(StageBase):
    #         output = FlowDataBase

    # except NotImplementedError as e:
    #     print(e)

    # else:
    #     raise Exception("Should have raised NotImplementedError")

    # @dataclass
    # class MockInputOutput(FlowDataBase):
    #     text: str

    # class StageTest(StageBase):
    #     input = MockInputOutput
    #     output = MockInputOutput

    #     def handle_data(self, data):
    #         return self.output(text=data["text"] + "processed")

    # result = StageTest().process(MockInputOutput(text="Hello world!"))
    # print(result.text)

    # pipeline = StageTest(next_stage=StageTest(next_stage=StageTest()))
    # result = pipeline.process(MockInputOutput(text="Hello world!"))
    # print(result.text)

    # class BadMockInput(FlowDataBase):
    #     text: str
    #     other: str

    # class StageBad(StageBase):
    #     input = BadMockInput
    #     output = BadMockInput

    #     def handle_data(self, data):
    #         return self.output(text=data["text"] + "processed")

    # try:
    #     pipeline = StageTest(next_stage=StageBad(next_stage=StageTest()))
    # except ExceptionPipesDoNotFit as e:
    #     print(e)



class PromptError(Exception):
    pass


class TemplateRenderError(PromptError):
    "TODO: Move to exceptions file"
    pass


class PromptBadQuestionTypeError(PromptError):
    pass


class PromptBadLanguageModelTypeError(PromptError):
    pass



from pathlib import Path
import os
import platformdirs


import sys
import select


def get_input_with_timeout(prompt, timeout=5, default="y"):
    print(prompt, end="", flush=True)
    ready, _, _ = select.select([sys.stdin], [], [], timeout)
    if ready:
        return sys.stdin.readline().strip()
    print(f"\nNo input received within {timeout} seconds. Using default: {default}")
    return default


class ExpectedParrotKeyHandler:
    asked_to_store_file_name = "asked_to_store.txt"
    ep_key_file_name = "ep_api_key.txt"
    application_name = "edsl"

    @property
    def config_dir(self):
        return platformdirs.user_config_dir(self.application_name)

    def _ep_key_file_exists(self) -> bool:
        """Check if the Expected Parrot key file exists."""
        return Path(self.config_dir).joinpath(self.ep_key_file_name).exists()

    def ok_to_ask_to_store(self):
        """Check if it's okay to ask the user to store the key."""
        from edsl.config import CONFIG

        if CONFIG.get("EDSL_RUN_MODE") != "production":
            return False

        return (
            not Path(self.config_dir).joinpath(self.asked_to_store_file_name).exists()
        )

    def reset_asked_to_store(self):
        """Reset the flag that indicates whether the user has been asked to store the key."""
        asked_to_store_path = Path(self.config_dir).joinpath(
            self.asked_to_store_file_name
        )
        if asked_to_store_path.exists():
            os.remove(asked_to_store_path)
            print(
                "Deleted the file that indicates whether the user has been asked to store the key."
            )

    def ask_to_store(self, api_key) -> bool:
        """Ask the user if they want to store the Expected Parrot key. If they say "yes", store it."""
        if self.ok_to_ask_to_store():
            # can_we_store = get_input_with_timeout(
            #     "Would you like to store your Expected Parrot key for future use? (y/n): ",
            #     timeout=5,
            #     default="y",
            # )
            can_we_store = "y"
            if can_we_store.lower() == "y":
                Path(self.config_dir).mkdir(parents=True, exist_ok=True)
                self.store_ep_api_key(api_key)
                # print("Stored Expected Parrot API key at ", self.config_dir)
                return True
            else:
                Path(self.config_dir).mkdir(parents=True, exist_ok=True)
                with open(
                    Path(self.config_dir).joinpath(self.asked_to_store_file_name), "w"
                ) as f:
                    f.write("Yes")
        return False

    def get_ep_api_key(self):
        # check if the key is stored in the config_dir
        api_key = None
        api_key_from_cache = None
        api_key_from_os = None

        if self._ep_key_file_exists():
            with open(Path(self.config_dir).joinpath(self.ep_key_file_name), "r") as f:
                api_key_from_cache = f.read().strip()

        api_key_from_os = os.getenv("EXPECTED_PARROT_API_KEY")

        if api_key_from_os and api_key_from_cache:
            if api_key_from_os != api_key_from_cache:
                import warnings

                warnings.warn(
                    "WARNING: The Expected Parrot API key from the environment variable "
                    "differs from the one stored in the config directory. Using the one "
                    "from the environment variable."
                )
            api_key = api_key_from_os

        if api_key_from_os and not api_key_from_cache:
            api_key = api_key_from_os

        if not api_key_from_os and api_key_from_cache:
            api_key = api_key_from_cache

        if api_key is not None:
            _ = self.ask_to_store(api_key)
        return api_key

    def delete_ep_api_key(self):
        key_path = Path(self.config_dir) / self.ep_key_file_name
        if key_path.exists():
            os.remove(key_path)
            print("Deleted Expected Parrot API key at ", key_path)

    def store_ep_api_key(self, api_key):
        # Create the directory if it doesn't exist
        os.makedirs(self.config_dir, exist_ok=True)

        # Create the path for the key file
        key_path = Path(self.config_dir) / self.ep_key_file_name

        # Save the key
        with open(key_path, "w") as f:
            f.write(api_key)
        # print("Stored Expected Parrot API key at ", key_path)



# """This module contains the Result class, which captures the result of one interview."""
from __future__ import annotations
import inspect
from collections import UserDict
from typing import Any, Type, Callable, Optional, TYPE_CHECKING, Union
from edsl.Base import Base
from edsl.utilities.decorators import add_edsl_version, remove_edsl_version

if TYPE_CHECKING:
    from edsl.agents.Agent import Agent
    from edsl.scenarios.Scenario import Scenario
    from edsl.language_models.LanguageModel import LanguageModel
    from edsl.prompts.Prompt import Prompt
    from edsl.surveys.Survey import Survey


QuestionName = str
AnswerValue = Any


class AgentNamer:
    """Maintains a registry of agent names to ensure unique naming."""

    def __init__(self):
        self._registry = {}

    def get_name(self, agent: "Agent") -> str:
        """Get or create a unique name for an agent."""
        agent_id = id(agent)
        if agent_id not in self._registry:
            self._registry[agent_id] = f"Agent_{len(self._registry)}"
        return self._registry[agent_id]


# Global instance for agent naming
agent_namer = AgentNamer().get_name


class Result(Base, UserDict):
    """
    This class captures the result of one interview.
    """

    def __init__(
        self,
        agent: "Agent",
        scenario: "Scenario",
        model: "LanguageModel",
        iteration: int,
        answer: dict[QuestionName, AnswerValue],
        prompt: dict[QuestionName, str] = None,
        raw_model_response: Optional[dict] = None,
        survey: Optional["Survey"] = None,
        question_to_attributes: Optional[dict[QuestionName, Any]] = None,
        generated_tokens: Optional[dict] = None,
        comments_dict: Optional[dict] = None,
        cache_used_dict: Optional[dict[QuestionName, bool]] = None,
        indices: Optional[dict] = None,
    ):
        """Initialize a Result object.

        :param agent: The Agent object.
        :param scenario: The Scenario object.
        :param model: The LanguageModel object.
        :param iteration: The iteration number.
        :param answer: The answer string.
        :param prompt: A dictionary of prompts.
        :param raw_model_response: The raw model response.
        :param survey: The Survey object.
        :param question_to_attributes: A dictionary of question attributes.
        :param generated_tokens: A dictionary of generated tokens.
        :param comments_dict: A dictionary of comments.
        :param cache_used_dict: A dictionary of cache usage.
        :param indices: A dictionary of indices.

        """
        self.question_to_attributes = (
            question_to_attributes or self._create_question_to_attributes(survey)
        )

        data = {
            "agent": agent,
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
            "prompt": prompt or {},
            "raw_model_response": raw_model_response or {},
            "question_to_attributes": question_to_attributes,
            "generated_tokens": generated_tokens or {},
            "comments_dict": comments_dict or {},
            "cache_used_dict": cache_used_dict or {},
        }
        super().__init__(**data)
        self.indices = indices
        self._sub_dicts = self._construct_sub_dicts()
        (
            self._combined_dict,
            self._problem_keys,
        ) = self._compute_combined_dict_and_problem_keys()

    @staticmethod
    def _create_question_to_attributes(survey):
        """Create a dictionary of question attributes."""
        if survey is None:
            return {}
        return {
            q.question_name: {
                "question_text": q.question_text,
                "question_type": q.question_type,
                "question_options": (
                    None if not hasattr(q, "question_options") else q.question_options
                ),
            }
            for q in survey.questions
        }

    @property
    def agent(self) -> "Agent":
        """Return the Agent object."""
        return self.data["agent"]

    @property
    def scenario(self) -> "Scenario":
        """Return the Scenario object."""
        return self.data["scenario"]

    @property
    def model(self) -> "LanguageModel":
        """Return the LanguageModel object."""
        return self.data["model"]

    @property
    def answer(self) -> dict[QuestionName, AnswerValue]:
        """Return the answers."""
        return self.data["answer"]

    @staticmethod
    def _create_agent_sub_dict(agent) -> dict:
        """Create a dictionary of agent details"""
        if agent.name is None:
            agent_name = agent_namer(agent)
        else:
            agent_name = agent.name

        return {
            "agent": agent.traits
            | {"agent_name": agent_name}
            | {"agent_instruction": agent.instruction},
        }

    @staticmethod
    def _create_model_sub_dict(model) -> dict:
        return {
            "model": model.parameters | {"model": model.model},
        }

    @staticmethod
    def _iteration_sub_dict(iteration) -> dict:
        return {
            "iteration": {"iteration": iteration},
        }

    def _construct_sub_dicts(self) -> dict[str, dict]:
        """Construct a dictionary of sub-dictionaries for the Result object."""
        sub_dicts_needing_new_keys = {
            "question_text": {},
            "question_options": {},
            "question_type": {},
        }

        for question_name in self.data["answer"]:
            if question_name in self.question_to_attributes:
                for dictionary_name in sub_dicts_needing_new_keys:
                    new_key = question_name + "_" + dictionary_name
                    sub_dicts_needing_new_keys[dictionary_name][new_key] = (
                        self.question_to_attributes[question_name][dictionary_name]
                    )

        new_cache_dict = {
            f"{k}_cache_used": v for k, v in self.data["cache_used_dict"].items()
        }

        d = {
            **self._create_agent_sub_dict(self.data["agent"]),
            **self._create_model_sub_dict(self.data["model"]),
            **self._iteration_sub_dict(self.data["iteration"]),
            "scenario": self.data["scenario"],
            "answer": self.data["answer"],
            "prompt": self.data["prompt"],
            "comment": self.data["comments_dict"],
            "generated_tokens": self.data["generated_tokens"],
            "raw_model_response": self.data["raw_model_response"],
            "question_text": sub_dicts_needing_new_keys["question_text"],
            "question_options": sub_dicts_needing_new_keys["question_options"],
            "question_type": sub_dicts_needing_new_keys["question_type"],
            "cache_used": new_cache_dict,
        }
        if hasattr(self, "indices") and self.indices is not None:
            d["agent"].update({"agent_index": self.indices["agent"]})
            d["scenario"].update({"scenario_index": self.indices["scenario"]})
            d["model"].update({"model_index": self.indices["model"]})
        return d

    @property
    def sub_dicts(self) -> dict[str, dict]:
        """Return a dictionary where keys are strings for each of the main class attributes/objects."""
        if self._sub_dicts is None:
            self._sub_dicts = self._construct_sub_dicts()
        return self._sub_dicts

    def check_expression(self, expression: str) -> None:
        for key in self.problem_keys:
            if key in expression and not key + "." in expression:
                raise ValueError(
                    f"Key by iself {key} is problematic. Use the full key {key + '.' + key} name instead."
                )
        return None

    def code(self):
        """Return a string of code that can be used to recreate the Result object."""
        raise NotImplementedError

    @property
    def problem_keys(self) -> list[str]:
        """Return a list of keys that are problematic."""
        return self._problem_keys

    def _compute_combined_dict_and_problem_keys(
        self,
    ) -> tuple[dict[str, Any], list[str]]:
        combined = {}
        problem_keys = []
        for key, sub_dict in self.sub_dicts.items():
            combined.update(sub_dict)
            # in some cases, the sub_dict might have keys that conflict with the main dict
            if key in combined:
                # The key is already in the combined dict
                problem_keys = problem_keys + [key]

            combined.update({key: sub_dict})
            # I *think* this allows us to do do things like "answer.how_feelling" i.e., that the evaluator can use
            # dot notation to access the subdicts.
        return combined, problem_keys

    @property
    def combined_dict(self) -> dict[str, Any]:
        """Return a dictionary that includes all sub_dicts, but also puts the key-value pairs in each sub_dict as a key_value pair in the combined dictionary.

        >>> r = Result.example()
        >>> r.combined_dict['how_feeling']
        'OK'
        """
        if self._combined_dict is None or self._problem_keys is None:
            (
                self._combined_dict,
                self._problem_keys,
            ) = self._compute_combined_dict_and_problem_keys()
        return self._combined_dict

    @property
    def problem_keys(self) -> list[str]:
        """Return a list of keys that are problematic."""
        if self._combined_dict is None or self._problem_keys is None:
            self._compute_combined_dict_and_problem_keys()
        return self._problem_keys

    def get_value(self, data_type: str, key: str) -> Any:
        """Return the value for a given data type and key.

        >>> r = Result.example()
        >>> r.get_value("answer", "how_feeling")
        'OK'

        - data types can be "agent", "scenario", "model", or "answer"
        - keys are relevant attributes of the Objects the data types represent
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self) -> dict[str, str]:
        """Return a dictionary where keys are object attributes and values are the data type (object) that the attribute is associated with.

        >>> r = Result.example()
        >>> r.key_to_data_type["how_feeling"]
        'answer'

        """
        d = {}
        problem_keys = []
        data_types = sorted(self.sub_dicts.keys())
        for data_type in data_types:
            for key in self.sub_dicts[data_type]:
                if key in d:
                    import warnings

                    warnings.warn(
                        f"Key '{key}' of data type '{data_type}' is already in use. Renaming to {key}_{data_type}"
                    )
                    problem_keys.append((key, data_type))
                    key = f"{key}_{data_type}"
                d[key] = data_type

        for key, data_type in problem_keys:
            self.sub_dicts[data_type][f"{key}_{data_type}"] = self.sub_dicts[
                data_type
            ].pop(key)
        return d

    def copy(self) -> Result:
        """Return a copy of the Result object.

        >>> r = Result.example()
        >>> r2 = r.copy()
        >>> r == r2
        True
        >>> id(r) == id(r2)
        False
        """
        return Result.from_dict(self.to_dict())

    def __eq__(self, other) -> bool:
        """Return True if the Result object is equal to another Result object.

        >>> r = Result.example()
        >>> r == r
        True

        """
        return hash(self) == hash(other)

    def to_dict(
        self, add_edsl_version: bool = True, include_cache_info: bool = False
    ) -> dict[str, Any]:
        """Return a dictionary representation of the Result object.

        >>> r = Result.example()
        >>> r.to_dict()['scenario']
        {'period': 'morning', 'scenario_index': 0, 'edsl_version': '...', 'edsl_class_name': 'Scenario'}
        """

        def convert_value(value, add_edsl_version=True):
            if hasattr(value, "to_dict"):
                return value.to_dict(add_edsl_version=add_edsl_version)
            else:
                return value

        d = {}
        for key, value in self.items():
            d[key] = convert_value(value, add_edsl_version=add_edsl_version)

            if key == "prompt":
                new_prompt_dict = {}
                for prompt_name, prompt_obj in value.items():
                    new_prompt_dict[prompt_name] = (
                        prompt_obj
                        if not hasattr(prompt_obj, "to_dict")
                        else prompt_obj.to_dict()
                    )
                d[key] = new_prompt_dict
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "Result"

        if include_cache_info:
            d["cache_used_dict"] = self.data["cache_used_dict"]
        else:
            d.pop("cache_used_dict", None)

        return d

    def __hash__(self):
        """Return a hash of the Result object."""
        from edsl.utilities.utilities import dict_hash

        return dict_hash(self.to_dict(add_edsl_version=False, include_cache_info=False))

    @classmethod
    @remove_edsl_version
    def from_dict(self, json_dict: dict) -> Result:
        """Return a Result object from a dictionary representation."""

        from edsl.agents.Agent import Agent
        from edsl.scenarios.Scenario import Scenario
        from edsl.language_models.LanguageModel import LanguageModel
        from edsl.prompts.Prompt import Prompt

        prompt_data = json_dict.get("prompt", {})
        prompt_d = {}
        for prompt_name, prompt_obj in prompt_data.items():
            prompt_d[prompt_name] = Prompt.from_dict(prompt_obj)

        result = Result(
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
            prompt=prompt_d,  # json_dict["prompt"],
            raw_model_response=json_dict.get(
                "raw_model_response", {"raw_model_response": "No raw model response"}
            ),
            question_to_attributes=json_dict.get("question_to_attributes", None),
            generated_tokens=json_dict.get("generated_tokens", {}),
            comments_dict=json_dict.get("comments_dict", {}),
            cache_used_dict=json_dict.get("cache_used_dict", {}),
        )
        return result

    def __repr__(self):
        """Return a string representation of the Result object."""
        params = ", ".join(f"{key}={repr(value)}" for key, value in self.data.items())
        return f"{self.__class__.__name__}({params})"

    @classmethod
    def example(cls):
        """Return an example Result object.

        >>> Result.example()
        Result(...)

        """
        from edsl.results.Results import Results

        return Results.example()[0]

    def score(self, scoring_function: Callable) -> Union[int, float]:
        """Score the result using a passed-in scoring function.

        >>> def f(status): return 1 if status == 'Joyful' else 0
        >>> Result.example().score(f)
        1
        """
        signature = inspect.signature(scoring_function)
        params = {}
        for k, v in signature.parameters.items():
            if k in self.combined_dict:
                params[k] = self.combined_dict[k]
            elif v.default is not v.empty:
                params[k] = v.default
            else:
                raise ValueError(f"Parameter {k} not found in Result object")
        return scoring_function(**params)

    @classmethod
    def from_interview(
        cls, interview, extracted_answers, model_response_objects
    ) -> Result:
        """Return a Result object from an interview dictionary."""

        def get_question_results(
            model_response_objects,
        ) -> dict[str, "EDSLResultObjectInput"]:
            """Maps the question name to the EDSLResultObjectInput."""
            question_results = {}
            for result in model_response_objects:
                question_results[result.question_name] = result
            return question_results

        def get_generated_tokens_dict(answer_key_names) -> dict[str, str]:
            generated_tokens_dict = {
                k + "_generated_tokens": question_results[k].generated_tokens
                for k in answer_key_names
            }
            return generated_tokens_dict

        def get_comments_dict(answer_key_names) -> dict[str, str]:
            comments_dict = {
                k + "_comment": question_results[k].comment for k in answer_key_names
            }
            return comments_dict

        def get_question_name_to_prompts(
            model_response_objects,
        ) -> dict[str, dict[str, str]]:
            question_name_to_prompts = dict({})
            for result in model_response_objects:
                question_name = result.question_name
                question_name_to_prompts[question_name] = {
                    "user_prompt": result.prompts["user_prompt"],
                    "system_prompt": result.prompts["system_prompt"],
                }
            return question_name_to_prompts

        def get_prompt_dictionary(answer_key_names, question_name_to_prompts):
            prompt_dictionary = {}
            for answer_key_name in answer_key_names:
                prompt_dictionary[answer_key_name + "_user_prompt"] = (
                    question_name_to_prompts[answer_key_name]["user_prompt"]
                )
                prompt_dictionary[answer_key_name + "_system_prompt"] = (
                    question_name_to_prompts[answer_key_name]["system_prompt"]
                )
            return prompt_dictionary

        def get_raw_model_results_and_cache_used_dictionary(model_response_objects):
            raw_model_results_dictionary = {}
            cache_used_dictionary = {}
            for result in model_response_objects:
                question_name = result.question_name
                raw_model_results_dictionary[question_name + "_raw_model_response"] = (
                    result.raw_model_response
                )
                raw_model_results_dictionary[question_name + "_cost"] = result.cost
                one_use_buys = (
                    "NA"
                    if isinstance(result.cost, str)
                    or result.cost == 0
                    or result.cost is None
                    else 1.0 / result.cost
                )
                raw_model_results_dictionary[question_name + "_one_usd_buys"] = (
                    one_use_buys
                )
                cache_used_dictionary[question_name] = result.cache_used

            return raw_model_results_dictionary, cache_used_dictionary

        question_results = get_question_results(model_response_objects)
        answer_key_names = list(question_results.keys())
        generated_tokens_dict = get_generated_tokens_dict(answer_key_names)
        comments_dict = get_comments_dict(answer_key_names)
        answer_dict = {k: extracted_answers[k] for k in answer_key_names}

        question_name_to_prompts = get_question_name_to_prompts(model_response_objects)
        prompt_dictionary = get_prompt_dictionary(
            answer_key_names, question_name_to_prompts
        )
        raw_model_results_dictionary, cache_used_dictionary = (
            get_raw_model_results_and_cache_used_dictionary(model_response_objects)
        )

        result = cls(
            agent=interview.agent,
            scenario=interview.scenario,
            model=interview.model,
            iteration=interview.iteration,
            # Computed objects
            answer=answer_dict,
            prompt=prompt_dictionary,
            raw_model_response=raw_model_results_dictionary,
            survey=interview.survey,
            generated_tokens=generated_tokens_dict,
            comments_dict=comments_dict,
            cache_used_dict=cache_used_dictionary,
            indices=interview.indices,
        )
        result.interview_hash = interview.initial_hash
        return result


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import json
import html
import re

from edsl import Question
from edsl import Survey

qualtrics_codes = {
    "TE": "free_text",
    "MC": "multiple_choice",
    "Matrix": "matrix",
    "DB": "instruction",  # not quite right, but for now
    "Timing": "free_text",  # not quite right, but for now
}
# TE (Text Entry): Allows respondents to input a text response.
# MC (Multiple Choice): Provides respondents with a list of options to choose from.
# DB (Descriptive Text or Information): Displays text or information without requiring a response.
# Matrix: A grid-style question where respondents can evaluate multiple items using the same set of response options.


def clean_html(raw_html):
    # Unescape HTML entities
    clean_text = html.unescape(raw_html)
    # Remove HTML tags
    clean_text = re.sub(r"<.*?>", "", clean_text)
    # Replace non-breaking spaces with regular spaces
    clean_text = clean_text.replace("\xa0", " ")
    # Optionally, strip leading/trailing spaces
    clean_text = clean_text.strip()
    return clean_text


class QualtricsQuestion:
    def __init__(self, question_json, debug=False):
        self.debug = debug
        self.question_json = question_json
        if self.element != "SQ":
            raise ValueError("Invalid question element type")

    @property
    def element(self):
        return self.question_json["Element"]

    @property
    def selector(self):
        return self.question_json.get("Selector", None)

    @property
    def question_name(self):
        return self.question_json["PrimaryAttribute"]

    @property
    def question_text(self):
        return clean_html(self.question_json["Payload"]["QuestionText"])

    @property
    def raw_question_type(self):
        return self.question_json["Payload"]["QuestionType"]

    @property
    def question_type(self):
        q_type = qualtrics_codes.get(self.raw_question_type, None)
        if q_type is None:
            print(f"Unknown question type: {self.raw_question_type}")
            return None
        return q_type

    @property
    def choices(self):
        if "Choices" in self.question_json["Payload"]:
            return [
                choice["Display"]
                for choice in self.question_json["Payload"]["Choices"].values()
            ]
        return None

    @property
    def answers(self):
        if "Answers" in self.question_json["Payload"]:
            return [
                choice["Display"]
                for choice in self.question_json["Payload"]["Choices"].values()
            ]
        return None

    def to_edsl(self):
        if self.question_type == "instruction":
            from edsl import Instruction

            return [Instruction(text=self.question_text, name=self.question_name)]

        if self.question_type == "free_text":
            try:
                q = Question(
                    **{
                        "question_type": self.question_type,
                        "question_text": self.question_text,
                        "question_name": self.question_name,
                    }
                )
                return [q]
            except Exception as e:
                return []

        if self.question_type == "multiple_choice":
            # Let's figure of it it's actually a checkbox question
            if self.selector == "MAVR" or self.selector == "MULTIPLE":
                try:
                    q = Question(
                        **{
                            "question_type": "checkbox",
                            "question_text": self.question_text,
                            "question_name": self.question_name,
                            "question_options": self.choices,
                        }
                    )
                    return [q]
                except Exception as e:
                    return []

            # maybe it's a linear scale!
            if "<br>" in self.choices[0]:
                option_labels = {}
                question_options = []
                for choice in self.choices:
                    if "<br>" in choice:
                        option_label, question_option = choice.split("<br>")
                        option_labels[int(question_option)] = option_label
                        question_options.append(int(question_option))
                    else:
                        question_options.append(int(choice))
                try:
                    q = Question(
                        **{
                            "question_type": "linear_scale",
                            "question_text": self.question_text,
                            "question_name": self.question_name,
                            "question_options": question_options,
                            "option_labels": option_labels,
                        }
                    )
                    return [q]
                except Exception as e:
                    if self.debug:
                        raise e
                    else:
                        print(e)
                        return []

            try:
                q = Question(
                    **{
                        "question_type": self.question_type,
                        "question_text": self.question_text,
                        "question_name": self.question_name,
                        "question_options": self.choices,
                    }
                )
                return [q]
            except Exception as e:
                return []

        if self.question_type == "matrix":
            questions = []
            for index, choice in enumerate(self.choices):
                try:
                    q = Question(
                        **{
                            "question_type": "multiple_choice",
                            "question_text": self.question_text + f" ({choice})",
                            "question_name": self.question_name + f"_{index}",
                            "question_options": self.answers,
                        }
                    )
                    questions.append(q)
                except Exception as e:
                    continue

            return questions

        raise ValueError(f"Invalid question type: {self.question_type}")


class SurveyQualtricsImport:
    def __init__(self, qsf_file_name: str):
        self.qsf_file_name = qsf_file_name
        self.question_data = self.extract_questions_from_json()

    def create_survey(self):
        questions = []
        for qualtrics_questions in self.question_data:
            questions.extend(qualtrics_questions.to_edsl())
        return Survey(questions)

    @property
    def survey_data(self):
        with open(self.qsf_file_name, "r") as f:
            survey_data = json.load(f)
        return survey_data

    def extract_questions_from_json(self):
        questions = self.survey_data["SurveyElements"]

        extracted_questions = []

        for question in questions:
            if question["Element"] == "SQ":
                extracted_questions.append(QualtricsQuestion(question))

        return extracted_questions

    def extract_blocks_from_json(self):
        blocks = []

        for element in self.survey_data["SurveyElements"]:
            if element["Element"] == "BL":
                for block_payload in element["Payload"]:
                    block_elements = [
                        BlockElement(be["Type"], be["QuestionID"])
                        for be in block_payload["BlockElements"]
                    ]
                    options_data = block_payload.get("Options", {})
                    options = BlockOptions(
                        options_data.get("BlockLocking", "false"),
                        options_data.get("RandomizeQuestions", "false"),
                        options_data.get("BlockVisibility", "Collapsed"),
                    )

                    block = SurveyBlock(
                        block_payload["Type"],
                        block_payload["Description"],
                        block_payload["ID"],
                        block_elements,
                        options,
                    )
                    blocks.append(block)

        return blocks


class SurveyBlock:
    def __init__(self, block_type, description, block_id, block_elements, options):
        self.block_type = block_type
        self.description = description
        self.block_id = block_id
        self.block_elements = block_elements
        self.options = options

    def __repr__(self):
        return f"SurveyBlock(type={self.block_type}, description={self.description}, id={self.block_id})"


class BlockElement:
    def __init__(self, element_type, question_id):
        self.element_type = element_type
        self.question_id = question_id

    def __repr__(self):
        return f"BlockElement(type={self.element_type}, question_id={self.question_id})"


class BlockOptions:
    def __init__(self, block_locking, randomize_questions, block_visibility):
        self.block_locking = block_locking
        self.randomize_questions = randomize_questions
        self.block_visibility = block_visibility

    def __repr__(self):
        return (
            f"BlockOptions(block_locking={self.block_locking}, "
            f"randomize_questions={self.randomize_questions}, "
            f"block_visibility={self.block_visibility})"
        )


if __name__ == "__main__":
    survey_creator = SurveyQualtricsImport("example.qsf")
    # print(survey_creator.question_data)
    # survey = survey_creator.create_survey()
    # info = survey.push()
    # print(info)
    # questions = survey.extract_questions_from_json()
    # for question in questions:
    #    print(question)



from collections import UserDict
from dataclasses import asdict
from edsl.enums import service_to_api_keyname

from edsl.language_models.key_management.models import LanguageModelInput


class KeyLookup(UserDict):
    """A class for looking up API keys and related configuration.

    >>> from edsl.language_models.key_management.models import LanguageModelInput
    >>> lookup = KeyLookup()
    >>> lm_input = LanguageModelInput.example()
    >>> lookup['test'] = lm_input
    >>> lookup.to_dict()['test']['api_token']
    'sk-abcd123'
    >>> restored = KeyLookup.from_dict(lookup.to_dict())
    >>> restored['test'].api_token
    'sk-abcd123'
    """

    def to_dict(self):
        """
        >>> kl = KeyLookup.example()
        >>> kl2 = KeyLookup.from_dict(kl.to_dict())
        >>> kl2 == kl
        True
        >>> kl2 is kl
        False
        """
        return {k: asdict(v) for k, v in self.data.items()}

    @classmethod
    def from_dict(cls, d):
        return cls({k: LanguageModelInput(**v) for k, v in d.items()})

    @classmethod
    def example(cls):
        return cls(
            {
                "test": LanguageModelInput.example(),
                "openai": LanguageModelInput.example(),
            }
        )

    def to_dot_env(self):
        """Return a string representation of the key lookup collection for a .env file."""
        lines = []
        for service, lm_input in self.items():
            if service != "test":
                lines.append(f"EDSL_SERVICE_RPM_{service.upper()}={lm_input.rpm}")
                lines.append(f"EDSL_SERVICE_TPM_{service.upper()}={lm_input.tpm}")
                key_name = service_to_api_keyname.get(service, service)
                lines.append(f"{key_name.upper()}={lm_input.api_token}")
                if lm_input.api_id is not None:
                    lines.append(f"{service.upper()}_API_ID={lm_input.api_id}")
        return "\n".join([f"{line}" for line in lines])


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
